@@ -16,13 +16,13 @@
 #include "soup-auth.h"
 #include "soup-auth-basic.h"
 #include "soup-auth-digest.h"
-#include "soup-auth-manager.h"
 #include "soup-auth-manager-ntlm.h"
 #include "soup-connection.h"
 #include "soup-marshal.h"
 #include "soup-message-private.h"
 #include "soup-message-queue.h"
 #include "soup-session.h"
+#include "soup-session-feature.h"
 #include "soup-session-private.h"
 #include "soup-socket.h"
 #include "soup-ssl.h"
@@ -75,8 +75,8 @@ typedef struct {
 
 	char *user_agent;
 
+	GSList *features;
 	SoupAuthManager *auth_manager;
-	SoupAuthManagerNTLM *ntlm_manager;
 
 	GHashTable *hosts; /* SoupURI -> SoupSessionHost */
 	GHashTable *conns; /* SoupConnection -> SoupSessionHost */
@@ -102,6 +102,10 @@ static void requeue_message (SoupSession *session, SoupMessage *msg);
 static void cancel_message  (SoupSession *session, SoupMessage *msg,
 			     guint status_code);
 
+static void auth_manager_authenticate (SoupAuthManager *manager,
+				       SoupMessage *msg, SoupAuth *auth,
+				       gboolean retrying, gpointer user_data);
+
 /* temporary until we fix this to index hosts by SoupAddress */
 extern guint     soup_uri_host_hash  (gconstpointer  key);
 extern gboolean  soup_uri_host_equal (gconstpointer  v1,
@@ -137,6 +141,9 @@ enum {
 	PROP_TIMEOUT,
 	PROP_USER_AGENT,
 	PROP_IDLE_TIMEOUT,
+	PROP_ADD_FEATURE,
+	PROP_ADD_FEATURE_BY_TYPE,
+	PROP_REMOVE_FEATURE_BY_TYPE,
 
 	LAST_PROP
 };
@@ -161,9 +168,14 @@ soup_session_init (SoupSession *session)
 	priv->max_conns = SOUP_SESSION_MAX_CONNS_DEFAULT;
 	priv->max_conns_per_host = SOUP_SESSION_MAX_CONNS_PER_HOST_DEFAULT;
 
-	priv->auth_manager = soup_auth_manager_new (session);
+	priv->auth_manager = g_object_new (SOUP_TYPE_AUTH_MANAGER_NTLM,
+					   SOUP_AUTH_MANAGER_NTLM_USE_NTLM, FALSE,
+					   NULL);
+	g_signal_connect (priv->auth_manager, "authenticate",
+			  G_CALLBACK (auth_manager_authenticate), session);
 	soup_auth_manager_add_type (priv->auth_manager, SOUP_TYPE_AUTH_BASIC);
 	soup_auth_manager_add_type (priv->auth_manager, SOUP_TYPE_AUTH_DIGEST);
+	soup_session_add_feature (session, SOUP_SESSION_FEATURE (priv->auth_manager));
 }
 
 static gboolean
@@ -197,6 +209,9 @@ dispose (GObject *object)
 	soup_session_abort (session);
 	cleanup_hosts (priv);
 
+	while (priv->features)
+		soup_session_remove_feature (session, priv->features->data);
+
 	G_OBJECT_CLASS (soup_session_parent_class)->dispose (object);
 }
 
@@ -214,9 +229,8 @@ finalize (GObject *object)
 
 	g_free (priv->user_agent);
 
-	soup_auth_manager_free (priv->auth_manager);
-	if (priv->ntlm_manager)
-		soup_auth_manager_ntlm_free (priv->ntlm_manager);
+	if (priv->auth_manager)
+		g_object_unref (priv->auth_manager);
 
 	if (priv->proxy_uri)
 		soup_uri_free (priv->proxy_uri);
@@ -472,6 +486,28 @@ soup_session_class_init (SoupSessionClass *session_class)
 				     "User-Agent string",
 				     NULL,
 				     G_PARAM_READWRITE));
+
+	g_object_class_install_property (
+		object_class, PROP_ADD_FEATURE,
+		g_param_spec_object (SOUP_SESSION_ADD_FEATURE,
+				     "Add Feature",
+				     "Add a feature object to the session",
+				     SOUP_TYPE_SESSION_FEATURE,
+				     G_PARAM_READWRITE));
+	g_object_class_install_property (
+		object_class, PROP_ADD_FEATURE_BY_TYPE,
+		g_param_spec_gtype (SOUP_SESSION_ADD_FEATURE_BY_TYPE,
+				    "Add Feature By Type",
+				    "Add a feature object of the given type to the session",
+				    SOUP_TYPE_SESSION_FEATURE,
+				    G_PARAM_READWRITE));
+	g_object_class_install_property (
+		object_class, PROP_REMOVE_FEATURE_BY_TYPE,
+		g_param_spec_gtype (SOUP_SESSION_REMOVE_FEATURE_BY_TYPE,
+				    "Remove Feature By Type",
+				    "Remove features of the given type from the session",
+				    SOUP_TYPE_SESSION_FEATURE,
+				    G_PARAM_READWRITE));
 }
 
 static gboolean
@@ -534,15 +570,9 @@ set_property (GObject *object, guint prop_id,
 		priv->max_conns_per_host = g_value_get_int (value);
 		break;
 	case PROP_USE_NTLM:
-		if (g_value_get_boolean (value)) {
-			if (!priv->ntlm_manager)
-				priv->ntlm_manager = soup_auth_manager_ntlm_new (session);
-		} else {
-			if (priv->ntlm_manager) {
-				soup_auth_manager_ntlm_free (priv->ntlm_manager);
-				priv->ntlm_manager = NULL;
-			}
-		}
+		g_object_set_property (G_OBJECT (priv->auth_manager),
+				       SOUP_AUTH_MANAGER_NTLM_USE_NTLM,
+				       value);
 		break;
 	case PROP_SSL_CA_FILE:
 		new_ca_file = g_value_get_string (value);
@@ -589,6 +619,15 @@ set_property (GObject *object, guint prop_id,
 	case PROP_IDLE_TIMEOUT:
 		priv->idle_timeout = g_value_get_uint (value);
 		break;
+	case PROP_ADD_FEATURE:
+		soup_session_add_feature (session, g_value_get_object (value));
+		break;
+	case PROP_ADD_FEATURE_BY_TYPE:
+		soup_session_add_feature_by_type (session, g_value_get_gtype (value));
+		break;
+	case PROP_REMOVE_FEATURE_BY_TYPE:
+		soup_session_remove_feature_by_type (session, g_value_get_gtype (value));
+		break;
 	default:
 		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
 		break;
@@ -613,7 +652,9 @@ get_property (GObject *object, guint prop_id,
 		g_value_set_int (value, priv->max_conns_per_host);
 		break;
 	case PROP_USE_NTLM:
-		g_value_set_boolean (value, priv->ntlm_manager != NULL);
+		g_object_get_property (G_OBJECT (priv->auth_manager),
+				       SOUP_AUTH_MANAGER_NTLM_USE_NTLM,
+				       value);
 		break;
 	case PROP_SSL_CA_FILE:
 		g_value_set_string (value, priv->ssl_ca_file);
@@ -713,13 +754,19 @@ free_host (SoupSessionHost *host)
 	g_slice_free (SoupSessionHost, host);
 }	
 
-void
-soup_session_emit_authenticate (SoupSession *session, SoupMessage *msg,
-				SoupAuth *auth, gboolean retrying)
+static void
+auth_manager_authenticate (SoupAuthManager *manager, SoupMessage *msg,
+			   SoupAuth *auth, gboolean retrying,
+			   gpointer session)
 {
 	g_signal_emit (session, signals[AUTHENTICATE], 0, msg, auth, retrying);
 }
 
+#define SOUP_METHOD_IS_SAFE(method) (method == SOUP_METHOD_GET || \
+				     method == SOUP_METHOD_HEAD || \
+				     method == SOUP_METHOD_OPTIONS || \
+				     method == SOUP_METHOD_PROPFIND)
+
 static void
 redirect_handler (SoupMessage *msg, gpointer user_data)
 {
@@ -730,16 +777,9 @@ redirect_handler (SoupMessage *msg, gpointer user_data)
 	new_loc = soup_message_headers_get (msg->response_headers, "Location");
 	g_return_if_fail (new_loc != NULL);
 
-	if (msg->status_code == SOUP_STATUS_MOVED_PERMANENTLY ||
-	    msg->status_code == SOUP_STATUS_TEMPORARY_REDIRECT) {
-		/* Don't redirect non-safe methods */
-		if (msg->method != SOUP_METHOD_GET &&
-		    msg->method != SOUP_METHOD_HEAD &&
-		    msg->method != SOUP_METHOD_OPTIONS &&
-		    msg->method != SOUP_METHOD_PROPFIND)
-			return;
-	} else if (msg->status_code == SOUP_STATUS_SEE_OTHER ||
-		   msg->status_code == SOUP_STATUS_FOUND) {
+	if (msg->status_code == SOUP_STATUS_SEE_OTHER ||
+	    (msg->status_code == SOUP_STATUS_FOUND &&
+	     !SOUP_METHOD_IS_SAFE (msg->method))) {
 		/* Redirect using a GET */
 		g_object_set (msg,
 			      SOUP_MESSAGE_METHOD, SOUP_METHOD_GET,
@@ -748,6 +788,12 @@ redirect_handler (SoupMessage *msg, gpointer user_data)
 					  SOUP_MEMORY_STATIC, NULL, 0);
 		soup_message_headers_set_encoding (msg->request_headers,
 						   SOUP_ENCODING_NONE);
+	} else if (msg->status_code == SOUP_STATUS_MOVED_PERMANENTLY ||
+		   msg->status_code == SOUP_STATUS_TEMPORARY_REDIRECT ||
+		   msg->status_code == SOUP_STATUS_FOUND) {
+		/* Don't redirect non-safe methods */
+		if (!SOUP_METHOD_IS_SAFE (msg->method))
+			return;
 	} else {
 		/* Three possibilities:
 		 *
@@ -914,6 +960,7 @@ connect_result (SoupConnection *conn, guint status, gpointer user_data)
 	 * any messages waiting for this host, since they're out
 	 * of luck.
 	 */
+	g_object_ref (session);
 	for (msg = soup_message_queue_first (priv->queue, &iter); msg; msg = soup_message_queue_next (priv->queue, &iter)) {
 		if (get_host_for_message (session, msg) == host) {
 			if (status == SOUP_STATUS_TRY_AGAIN) {
@@ -925,6 +972,7 @@ connect_result (SoupConnection *conn, guint status, gpointer user_data)
 			}
 		}
 	}
+	g_object_unref (session);
 }
 
 /**
@@ -1291,3 +1339,99 @@ soup_session_abort (SoupSession *session)
 
 	g_slist_free (conns);
 }
+
+/**
+ * soup_session_add_feature:
+ * @session: a #SoupSession
+ * @feature: an object that implements #SoupSessionFeature
+ *
+ * Adds @feature's functionality to @session. You can also add a
+ * feature to the session at construct time by using the
+ * %SOUP_SESSION_ADD_FEATURE property.
+ **/
+void
+soup_session_add_feature (SoupSession *session, SoupSessionFeature *feature)
+{
+	SoupSessionPrivate *priv;
+
+	g_return_if_fail (SOUP_IS_SESSION (session));
+	g_return_if_fail (SOUP_IS_SESSION_FEATURE (feature));
+
+	priv = SOUP_SESSION_GET_PRIVATE (session);
+	priv->features = g_slist_prepend (priv->features, g_object_ref (feature));
+	soup_session_feature_attach (feature, session);
+}
+
+/**
+ * soup_session_add_feature_by_type:
+ * @session: a #SoupSession
+ * @feature_type: the #GType of a class that implements #SoupSessionFeature
+ *
+ * Creates a new feature of type @feature_type and adds it to
+ * @session. You can use this instead of soup_session_add_feature() in
+ * the case wher you don't need to customize the new feature in any
+ * way. You can also add a feature to the session at construct time by
+ * using the %SOUP_SESSION_ADD_FEATURE_BY_TYPE property.
+ **/
+void
+soup_session_add_feature_by_type (SoupSession *session, GType feature_type)
+{
+	SoupSessionFeature *feature;
+
+	g_return_if_fail (SOUP_IS_SESSION (session));
+	g_return_if_fail (g_type_is_a (feature_type, SOUP_TYPE_SESSION_FEATURE));
+
+	feature = g_object_new (feature_type, NULL);
+	soup_session_add_feature (session, feature);
+	g_object_unref (feature);
+}
+
+/**
+ * soup_session_remove_feature:
+ * @session: a #SoupSession
+ * @feature: a feature that has previously been added to @session
+ *
+ * Removes @feature's functionality from @session.
+ **/
+void
+soup_session_remove_feature (SoupSession *session, SoupSessionFeature *feature)
+{
+	SoupSessionPrivate *priv;
+
+	g_return_if_fail (SOUP_IS_SESSION (session));
+
+	priv = SOUP_SESSION_GET_PRIVATE (session);
+	if (g_slist_find (priv->features, feature)) {
+		priv->features = g_slist_remove (priv->features, feature);
+		soup_session_feature_detach (feature, session);
+		g_object_unref (feature);
+	}
+}
+
+/**
+ * soup_session_remove_feature_by_type:
+ * @session: a #SoupSession
+ * @feature_type: the #GType of a class that implements #SoupSessionFeature
+ *
+ * Removes all features of type @feature_type (or any subclass of
+ * @feature_type) from @session. You can also remove standard features
+ * from the session at construct time by using the
+ * %SOUP_SESSION_REMOVE_FEATURE_BY_TYPE property.
+ **/
+void
+soup_session_remove_feature_by_type (SoupSession *session, GType feature_type)
+{
+	SoupSessionPrivate *priv;
+	GSList *f;
+
+	g_return_if_fail (SOUP_IS_SESSION (session));
+
+	priv = SOUP_SESSION_GET_PRIVATE (session);
+restart:
+	for (f = priv->features; f; f = f->next) {
+		if (G_TYPE_CHECK_INSTANCE_TYPE (f->data, feature_type)) {
+			soup_session_remove_feature (session, f->data);
+			goto restart;
+		}
+	}
+}
